import numpy as np
import pandas as pd
import numpy.core.multiarray
import fastcluster
from multiprocessing import Pool
from functools import partial


def flatten(container):
    for i in container:
        if isinstance(i, (list,tuple)):
            for j in flatten(i):
                yield j
        else:
            yield i
def flatx(x):
    return tuple(sorted(flatten(x)))

def AVdist(R):
    N = R.shape[0]
    d = R[np.triu_indices(N,1)]
    out = fastcluster.average(d)

    outI = out.astype(int)

    dend = {i:(i,) for i in xrange(N)}

    for i in xrange(len(outI)):
        dend[i+N] = (dend[outI[i][0]],dend[outI[i][1]])


    for i in xrange(N):
        dend.pop(i,None)

    dend = [(flatx(a),flatx(b)) for a,b in dend.values()]

    dend ={flatx((a,b)):(np.array(a),np.array(b)) for a,b in dend}
    return dend
   
def SingleBoot(X,nan,seed=None):
	local_state = np.random.RandomState(seed)
	sel = local_state.choice(range(X.shape[1]),replace=True,size=X.shape[1])
	Xb = X[:,sel]
	if nan==False:
		return 1.-np.corrcoef(Xb)
	else:
		return 1.-np.array(pd.DataFrame(Xb.T).corr())

    
def BootDist(X,Nt=1000,nan=False,ncpu=1):

	f = partial(SingleBoot,X,nan)
	seeds = np.random.randint(2**32,size=Nt)
	if ncpu==1:
		Rb = map(f,seeds)
	else:
		p = Pool(processes=ncpu)
		Rb = p.map(f,seeds)
		p.close()

	return Rb
    
def get_Pvalues(Rb,LV,N,alpha):
	vl = [(tuple(a),tuple(b)) for a,b in LV.values()]
	rxy = dict(zip(vl,map(lambda (a,b): np.array(map(lambda x:x[a][:,b].mean(),Rb)),LV.values())))

	Nt = float(len(Rb))
	PV = []
	for (a,b) in LV.values():
		a,b = tuple(a),tuple(b)
		t = b
		if LV.has_key(t):
			c,d = map(tuple,LV[t])
			PV.append((((rxy[(c,d)]-rxy[(a,b)])>0).sum()/Nt,t))
		t = a
		if LV.has_key(t):
			c,d = map(tuple,LV[t])
			PV.append((((rxy[(c,d)]-rxy[(a,b)])>0).sum()/Nt,t))
		
	PV = sorted(PV)
	
	L,PV = FDR(PV,alpha,N)
	
	return L,PV
	



def FDR(PV,alpha,N):
	p = np.array(zip(*PV)[0])
	thr = np.arange(1,len(PV)+1)*alpha/len(PV)

	sel = np.where(p<thr)[0]
	if len(sel)==0: 
		thr=-1.
	else:
		thr = thr[sel][-1]

	L = [range(N)]+[PV[i][1] for i in np.where(p<=thr)[0]]
	L = map(tuple,sorted(L,key=len,reverse=True))
	
	PV = {c:p for p,c in PV}
	PV[tuple(range(N))] = np.nan
	return L,PV

def Find_ValidatedCluster(X,Nt=1000,alpha=0.05,nan=False,ncpu=1):
	if nan==False:
		R = 1 - np.corrcoef(X)
	else:
		R = 1. -np.array(pd.DataFrame(X.T).corr()) 

	LV = AVdist(R)

	Rb = BootDist(X,Nt,nan,ncpu)
	
	L,PV = get_Pvalues(Rb,LV,X.shape[0],alpha)


	
	return L,PV,LV



from setuptools import setup

setup(	name='svhc',
      	version='0.12',
      	packages=['svhc'],
	install_requires=[	'numpy',
				'pandas',
				'fastcluster',
				'matplotlib',
				'scipy',
				'collections',
				'seaborn',
				'multiprocessing',
				'functools'],
	scripts=['bin/svhc','bin/svhc_benchmark','bin/svhc_plot'],
	author='Christian Bongiorno',
	author_email='pvofeta@gmail.com',
	license='GPL',
	zip_safe=False

      )




import matplotlib.pyplot as plt
import matplotlib.cm as cmap
from metric import StandardDendrogram,OrderArray,ToMemb
import matplotlib.patches as patches
import numpy as np
import seaborn as sns
import pandas as pd


def DendroAndCorrDist(x,L,LV,method='average',xylabel='Objects',file_w=None,color_style='Set1',nan=False):
    
   
	ordn = OrderArray(StandardDendrogram(ToMemb(LV.keys(),x.shape[0])))
	x = x[ordn]

	rev_ornd = dict(sorted(zip(ordn,range(len(ordn)))))

	LV = {tuple(sorted([rev_ornd[k] for k in K])):(np.array(sorted([rev_ornd[a] for a in LV[K][0]])),
												   np.array(sorted([rev_ornd[a] for a in LV[K][1]]))) for K in LV}

	L = [tuple(sorted([rev_ornd[k] for k in K])) for K in L]

	if nan==False:
		R = 1 - np.corrcoef(x)
	else:
		R = 1. -np.array(pd.DataFrame(x.T).corr()) 
	#R = 1.-np.corrcoef(x)

	L = sorted(L,key=len)

	N = R.shape[0]
	K = LV.keys()

	import matplotlib.patches as patches
	import colorsys
	nc  = len(L)
	import matplotlib as mpl
	import matplotlib.cm as cm

	cnmx = sns.color_palette(color_style, nc-1)

	from collections import defaultdict
	col = defaultdict(int)
	for k in K:
		for j in range(len(L)):
			if set(k) <= set(L[j]):

				if tuple(L[j])==tuple(range(N)):
					col[tuple(k)] = 'black'
					break

				col[tuple(k)] = cnmx[j-1]
				break


	from matplotlib import gridspec

	fig = plt.figure(figsize=(5,9))
	gs = gridspec.GridSpec(2, 1, height_ratios=[1, 1.34]) 

	# the fisrt subplot
	axarr = [0,0]
	axarr[0] = plt.subplot(gs[0])
	axarr[1] = plt.subplot(gs[1], sharex = axarr[0])


	pos = defaultdict(tuple)

	for i in range(N):
		pos[(i,)] = (float(i),0)

	K = sorted(LV.keys(),key=len)

	hm = []
	for k in K:
		d = tuple(sorted(LV[k][0])),tuple(sorted(LV[k][1]))

		c = tuple(sorted(k))
		#if not col.has_key(c):
		#    break
		pp = pos[d[0]][0],pos[d[1]][0]
		h0 = pos[d[0]][1],pos[d[1]][1]

		if method=='average':
			h = R[LV[k][0]][:,LV[k][1]].mean()
		elif method=='complete':
			h = R[LV[k][0]][:,LV[k][1]].min()
		elif method=='single':
			h = R[LV[k][0]][:,LV[k][1]].max()

		axarr[0].plot([pp[0],pp[0]],[h0[0],h],'-',color='k',alpha=0.5)
		axarr[0].plot([pp[1],pp[1]],[h0[1],h],'-',color='k',alpha=0.5)
		axarr[0].plot([pp[0],pp[1]],[h,h],'-',color='k',lw=1.5)
		axarr[0].plot([pp[0],pp[1]],[h,h],'-',color=col[c],lw=1.4,zorder=1000)

		hm.append(h)

		#plt.plot((pp[0]+pp[1])/2,h,'o',color=col[c],ms=4,zorder=1000)

		pos[c] = (pp[0]+pp[1])/2,h
		#print "ciao"

	axarr[0].set_yticks(np.arange(0,1.2,0.2))
	axarr[0].set_ylim([0,max(hm)+0.05])

	axarr[0].yaxis.tick_right()
	axarr[0].tick_params(labelsize=14)
	axarr[0].set_ylabel('Distance',fontsize=18)
	axarr[0].yaxis.set_label_position("right")

	if nan==False:
		R = np.corrcoef(x)
	else:
		R =  np.array(pd.DataFrame(x.T).corr()) 

	ims = axarr[1].imshow(R,vmin=-1,vmax=1,cmap=cm.RdBu_r,interpolation='nearest')


	cbar_ax = fig.add_axes([0.95, 0.133, 0.03, 0.4])
	cbr = fig.colorbar(ims, cax=cbar_ax)

	cbr.set_label('correlation',fontsize=18)
	axarr[1].set_xlabel(xylabel,fontsize=18)
	axarr[1].set_ylabel(xylabel,fontsize=18)
	axarr[1].tick_params(labelsize=14)

	axarr[1].set_ylim([0,N])

	for l in L:
		if l == tuple(range(N)): continue

		axarr[1].add_patch(
			patches.Rectangle(
				(min(l)-0.5, min(l)-0.5),
				len(l),
				len(l),
				fill=False,
				lw=1.,
				color='green',	
				)
			)

	plt.setp(axarr[0].get_xticklabels(), visible=False)
	plt.subplots_adjust(hspace=.0)


	plt.savefig(file_w+'.pdf',bbox_inches='tight')




import numpy as np
import igraph as ig
import sys
import pandas as pd
import hclustval


def ToMemb(nod,N):
    h = np.array([-1]*N)

    nod = sorted(nod,key=lambda x:len(x),reverse=True)
    Hn = []
    sel = {}
    while True:
        c = 0
        for i in range(len(nod)):
            if i in sel: continue
            if not all(h[np.array(nod[i])]==-1): continue

            h[np.array(nod[i])]=c
            sel[i] = None
            c+=1

        h[h==-1] = np.arange(h.max()+1,(h==-1).sum()+h.max()+1)
        Hn.append(h)
        h = np.array([-1]*N)
        if len(sel)==len(nod):
            break

    Hn.append(np.arange(N))    
    return Hn

def getPvalue(a,b):
    df =  ((a-b)>0).sum()
    if df>len(a)/2.:
        return (len(a)-df)/float(len(a))
    else:
        return df/float(len(a))

def srt(x):
	return tuple(sorted(x))

def AverageLink(R):
	N = R.shape[0]
	LV = {}

	nod =  [np.array([i]) for i in xrange(N)]

	for lv in xrange(N-1):
		Y = []
		for i in xrange(len(nod)):
			for j in xrange(i):
				Y.append((R[nod[i]][:,nod[j]].mean(),i,j))

		y,a,b = max(Y)
		
		sel = {a:None,b:None}
		
		LV[tuple(sorted(list(np.append(nod[a],nod[b]))))] = (nod[a],nod[b])

		nod = [nod[i] for i in range(len(nod)) if not sel.has_key(i)] + [np.append(nod[a],nod[b])]
	
	return LV

def SingleLink(R):
	N = R.shape[0]
	LV = {}

	nod =  [np.array([i]) for i in xrange(N)]

	for lv in xrange(N-1):
		Y = []
		for i in xrange(len(nod)):
			for j in xrange(i):
				Y.append((R[nod[i]][:,nod[j]].max(),i,j))

		y,a,b = max(Y)
		
		sel = {a:None,b:None}
		
		LV[tuple(sorted(list(np.append(nod[a],nod[b]))))] = (nod[a],nod[b])

		nod = [nod[i] for i in range(len(nod)) if not sel.has_key(i)] + [np.append(nod[a],nod[b])]
	
	return LV
	
def CompleteLink(R):
	N = R.shape[0]
	LV = {}

	nod =  [np.array([i]) for i in xrange(N)]

	for lv in xrange(N-1):
		Y = []
		for i in xrange(len(nod)):
			for j in xrange(i):
				Y.append((R[nod[i]][:,nod[j]].min(),i,j))

		y,a,b = max(Y)
		
		sel = {a:None,b:None}
		
		LV[tuple(sorted(list(np.append(nod[a],nod[b]))))] = (nod[a],nod[b])

		nod = [nod[i] for i in range(len(nod)) if not sel.has_key(i)] + [np.append(nod[a],nod[b])]
	
	return LV

def pairBoot(Rb):    
    Nt,N = len(Rb),Rb[0].shape[0]
    
    for i in xrange(N):
        for j in xrange(i):    
            ordn = np.random.choice(Nt,size=Nt,replace=False)
            cmd = [Rb[x][i,j] for x in xrange(Nt)]
            for x in xrange(Nt):
                Rb[x][i,j] = cmd[x]
                Rb[x][j,i] = cmd[x]
    return Rb	
	
def Boot(X,Nt=1000,boottype='standard',pattern=None,M=None):
	
	if pattern==None:
		Rb = []
		for i in xrange(Nt):
			sel = np.random.choice(range(X.shape[1]),replace=True,size=X.shape[1])
			Xb = X[:,sel]
			Rb.append(np.corrcoef(Xb))

		if boottype=='pair':
			Rb = pairBoot(Rb)
		return Rb
	else:
		Rb = []
		for _ in xrange(Nt):
			Xb,cmm = hclustval.Benchmark.CREATE_DATA(pattern,T=M)
			Rb.append(np.corrcoef(Xb))
			
		return Rb

def Collect(x,Rb,LV,method):
	
	if method=='average':
		d = [Rb[i][LV[srt(x)][0]][:,LV[srt(x)][1]].mean() for i in range(len(Rb))]
	elif method=='single':
		d = [Rb[i][LV[srt(x)][0]][:,LV[srt(x)][1]].max() for i in range(len(Rb))]
	elif method=='complete':
		d = [Rb[i][LV[srt(x)][0]][:,LV[srt(x)][1]].min() for i in range(len(Rb))]
		
	return np.array(d)


def Compare_Null(LV,Rb,method):
	C = {k:Collect(k,Rb,LV,method) for k in LV}

	P = []
	KEY = C.keys()
	for i in xrange(len(KEY)):
		for j in xrange(i):
			if len(np.intersect1d(KEY[i],KEY[j]))==0: continue
			if not (KEY[i] in map(tuple,map(sorted,LV[KEY[j]])) or KEY[j] in map(tuple,map(sorted,LV[KEY[i]]))): continue
			p= getPvalue(C[KEY[i]],C[KEY[j]])
			
			P.append((p,i,j))

	P = np.array(sorted(P))
	return KEY,P

def Find_Component(LV,Rb,method,alpha=0.05):
	
	KEY,P = Compare_Null(LV,Rb,method)
	indx = np.where(P[:,0]<alpha*np.arange(1,P.shape[0]+1)/P.shape[0])[0]

	if len(indx)==0:
		indx = 0
	else:
		indx = indx[-1]+1


	ed = list(map(tuple,P[indx:,1:].astype(int)))
	g = ig.Graph(n=len(KEY),edges=ed)
	g.vs["name"] = KEY

	comp = g.components()

	return comp


def HValidate(LV,Rb,method,alpha=0.05):
	comp = Find_Component(LV,Rb,method,alpha)

	KEY = comp.graph.vs["name"]
	comp = comp.membership
	L = [[] for i in range(len(set(comp)))]
	for i,c in enumerate(comp):
		L[c].extend(KEY[i])
	L = [list(set(l)) for l in L]

	return L
	
def HclustVal(X,alpha=0.05,Nt=1000,method='average',boottype='standard',Rb=None,pattern=None,M=None):
	
	if pattern!=None and M==None:
		print "If you specify the pattern, you must specify the dataseries lenght (M) too"
		sys.exit(0)
	
	
	R = np.corrcoef(X)
	
	if method!='all':
		if method=='average':
			LV = AverageLink(R)
		elif method=='single':
			LV = SingleLink(R)
		elif method=='complete':
			LV = CompleteLink(R)
		
		if Rb==None:
			Rb = Boot(X,Nt,boottype,pattern,M)
		
		L = HValidate(LV,Rb,method,alpha)
		
		return L,hclustval.StandardDendrogram(ToMemb(L,X.shape[0])),LV
	
	else:
		res = {}
		
		res['single'] = SingleLink(R)
		if Rb==None:
			Rb = Boot(X,Nt,boottype,pattern,M)
		res['singleVal'] = HValidate(res['single'],Rb,'single',alpha)
		#res['singValCompress'] = hclustval.StandardDendrogram(ToMemb(res['single'],X.shape[0]))
		
		res['average'] = AverageLink(R)
		if Rb==None:
			Rb = Boot(X,Nt,boottype,pattern,M)
		res['averageVal'] = HValidate(res['average'],Rb,'average',alpha)
		#res['averageValCompress'] = hclustval.StandardDendrogram(ToMemb(res['average'],X.shape[0]))
		
		res['complete'] = CompleteLink(R)
		if Rb==None:
			Rb = Boot(X,Nt,boottype,pattern,M)
		res['completeVal'] = HValidate(res['complete'],Rb,'complete',alpha)
		#res['completeValCompress'] = hclustval.StandardDendrogram(ToMemb(res['complete'],X.shape[0]))
		
		return res
		
		
	
if __name__=='__main__':
	
	inputfile,alpha,Nt,method = sys.argv[1:]
	
	alpha,Nt = float(alpha),int(Nt)
	
	X = np.array(pd.read_csv(inputfile,sep='\t',header=None))
	
	comm,H = HclustVal(X,alpha,Nt,method)
	
	with open('HyperNodeList_out.dat','w') as fw:
		fw.write("\n".join([",".join(map(str,comm[i])) for i in range(len(comm))]))

	pd.DataFrame(hclustval.CompressHC(comm,X.shape[0])).to_csv('CompressDendrogram_out.dat',sep='\t',index=False,header=False)


	
	
	
